

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Human Detector  async plugin for any platform """
import copy
import asyncio
import uuid
import logging
import os
import numpy as np
from threading import Thread
from foglamp.common import logger
from foglamp.plugins.common import utils
import async_ingest
import cv2
import time
import subprocess


_LOGGER = logger.setup(__name__, level=logging.INFO)
BACKGROUND_TASK = False


def check_background():
    out = subprocess.Popen(['systemctl', 'status', 'foglamp'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    stdout, stderr = out.communicate()
    if str(stdout).find("active (running)") != -1:
        return True
    else:
        return False


try:
    from tensorflow.lite.python.interpreter import Interpreter
    try:
        from tensorflow.lite.python.interpreter import load_delegate
    except ImportError:
        _LOGGER.exception("Edge TPU Support is not found in your tensorflow installation ")
        pass
except Exception as e:
    _LOGGER.exception("Tensoflow installation not found ")


__author__ = "Amandeep Singh Arora , Deepanshu Yadav"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Person Detection On FogLAMP',
        'type': 'string',
        'default': 'person_detection',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'detect.tflite',
        'order': '1',
        'displayName': 'TFlite model file'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'coco_labels.txt',
        'order': '2',
        'displayName': 'Labels file'
    },
    'asset_name': {
        'description': 'Asset Name',
        'type': 'string',
        'default': 'Detection Results',
        'order': '3',
        'displayName': 'Asset Name'
    },
    'enable Edge TPU': {
        'description': 'Connect the Coral Edge TPU and enable this',
        'type': 'boolean',
        'default': 'false',
        'order': '4',
        'displayName': 'Enable Edge TPU'
    },
    'Min Confidence Threshold': {
        'description': 'Threshold to select the detected objects',
        'type': 'float',
        'default': '0.5',
        'minimum':'0',
        'maximum':'1',
        'order': '5',
        'displayName': 'Minimum Confidence Threshold'
    },
    'Camera_ID': {
        'description': 'The number associated  with your video device. See /dev in your '
                       'filesystem you will see video0 or video1',
        'type': 'integer',
        'default': '0',
        'order': '6',
        'displayName': 'Camera ID'
    }

}


"""
Helper classs for starting a separate thread for reading frame from the Camera 
Device

"""


class VideoStream:
    """ Camera object that controls 
        video streaming from the Camera
    """
    def __init__(self, resolution=(640,480), framerate=30, source=0):
        # Initialize the PiCamera and the camera image stream

        self.stream = cv2.VideoCapture(source)
        ret = self.stream.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*'MJPG'))
        ret = self.stream.set(3, resolution[0])
        ret = self.stream.set(4, resolution[1])
            
        # Read first frame from the stream
        (self.grabbed, self.frame) = self.stream.read()

        # Variable to control when the camera is stopped
        self.stopped = False

    def start(self):

        # Start the thread that reads frames from the video stream
        Thread(target=self.update, args=(), name="Reader Thread").start()

        return self

    def update(self):
        # Keep looping indefinitely until the thread is stopped
        while True:
            # If the camera is stopped, stop the thread
            if self.stopped:

                # Release camera resources

                self.stream.release()

                return

            # Otherwise, grab the next frame from the stream
            (self.grabbed, self.frame) = self.stream.read()

    def read(self):
        # Return the most recent frame
        return self.frame

    def stop(self):
        # Indicate that the camera and thread should be stopped
        self.stopped = True


"""
    Helper Class for loading  the model and performing the inference 
    on it. 
"""


class Inference:
    def __init__(self):

        self.interpreter = None
        self.input_details = None
        self.output_details = None
        self.height_for_model = None
        self.width_for_model = None
        self.floating_model = None
        self.min_conf_threshold = None
        self.labels = None

    def get_interpreter(self, model, enable_tpu, labels, min_conf_threshold=0.5):
        """ Returns interpreter from the model
            Args: 
                  model -> full path of the .tflite file , 
                  enable_tpu-> whether you want to use Edge TPU or not
        Returns: 
                  interpreter object 
        Raises: 
                Raises exception if the Runtime Library for Edge TPU 
                is not found

        """
        if enable_tpu == 'true':
            try:     
                # loading the Edge TPU Runtime
                self.interpreter = Interpreter(model_path=model,
                                               experimental_delegates=[load_delegate('libedgetpu.so.1.0')])
                self.interpreter.allocate_tensors()
                self.input_details = self.interpreter.get_input_details()
                self.output_details = self.interpreter.get_output_details()
                self.height_for_model = self.input_details[0]['shape'][1]
                self.width_for_model = self.input_details[0]['shape'][2]
                self.floating_model = (self.input_details[0]['dtype'] == np.float32)
                self.min_conf_threshold = min_conf_threshold
                self.labels = labels
                return self.interpreter

            except Exception as e:
                _LOGGER.info('Failed to load edge TPU Runtime {}'.format(e))

        else:

            self.interpreter = Interpreter(model_path=model)

            self.interpreter.allocate_tensors()

            self.input_details = self.interpreter.get_input_details()
            self.output_details = self.interpreter.get_output_details()
            self.height_for_model = self.input_details[0]['shape'][1]
            self.width_for_model = self.input_details[0]['shape'][2]
            self.floating_model = (self.input_details[0]['dtype'] == np.float32)
            self.min_conf_threshold = min_conf_threshold
            self.labels = labels
            return self.interpreter

    def perform_inference(self, input_data):
        """ Returns bounding box , class , score
            Args: 
                   input_data ->  the input data to be fed into the model
            Returns: 
                     boxes -> an array of bounding box of the objects detected
                     classes-> an array of the class of objects detected
                     scores-> an array of scores(0 to 1) of the objects detected
            Raises: None
        """

        self.interpreter.set_tensor(self.input_details[0]['index'], input_data)
        self.interpreter.invoke()
        # Retrieve detection results

        # Bounding box coordinates of detected objects
        boxes = self.interpreter.get_tensor(self.output_details[0]['index'])[0] 

        # Class index of detected objects
        classes = self.interpreter.get_tensor(self.output_details[1]['index'])[0] 

        # Confidence of detected objects
        scores = self.interpreter.get_tensor(self.output_details[2]['index'])[0] 

        return boxes, classes, scores


# GLOBAL VARIABLES DECLARATION
c_callback = None
c_ingest_ref = None
loop = None
async_thread = None
camera_processing_thread = None
shutdown_in_progress = None
inference = None
asset_name = None


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Person Detection plugin',
        'version': '1.8.0',
        'mode': 'async',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    global shutdown_in_progress
    shutdown_in_progress = False
    global BACKGROUND_TASK
    BACKGROUND_TASK = check_background()
    data = copy.deepcopy(config)
    return data
   

def round_to_three_decimal_places(x):
    """ Rounds off a floating number to three decimal places
         Args: 
                x -> a floating number 
         Returns:
                returns a floating point number rounded off to three decimal places
        Raises: None
    """

    return round(x, 3)


def camera_loop(**kwargs):
    """ Main function that keeps on fetching frame , 
        performing inference and drawing the result of
        inference on the detection window.
         Args:
           Keyword Arguements -> Each one is listed below
         Returns:
           None 
         Raises:  
           Raises exception if unable to draw the frame on the the window.
    """

    # if floating point model is used we need to subtract the mean and divide 
    # by standard deviation
    input_mean = kwargs['input_mean']
    input_std = kwargs['input_std']

    # the height of the detection window on which frames are to be displayed
    camera_height = kwargs['camera_height']

    # the width of the detection window on which frames are to be displayed
    camera_width = kwargs['camera_width']

    # Note : The object contents may change if plugin_reconfigure is called 
    # thats why the inference object in the loop is global

    # these variables are used for calculation of frame per seconds (FPS)
    frame_rate_calc = 1
    freq = cv2.getTickFrequency()

    source = kwargs['source']

    # Initialize the stream object and start the thread that keeps on reading frames
    # This thread is independent of the Camera Processing Thread
    videostream = VideoStream(resolution=(camera_width, camera_height), source=source).start()

    # creating a window with a name
    window_name = 'Human detector'
    global BACKGROUND_TASK
    if not BACKGROUND_TASK:
        show_window = True
        cv2.namedWindow(window_name)
    else:
        show_window = False
   
    while True:
        # Capture frame-by-frame
        t1 = cv2.getTickCount()
        global inference, asset_name

        # we need the height , width to resize the image for feeding into the model
        height_for_model = inference.height_for_model
        width_for_model = inference.width_for_model

        #  check if floating point model is used or not
        floating_model = inference.floating_model

        # The minimum confidence to threshold the detections obtained from model
        min_conf_threshold = inference.min_conf_threshold

        # The list of labels of the supported objects detected by the plugin
        labels = inference.labels

        # Taking the frame the stream  
        frame1 = videostream.read()

        frame = frame1.copy()

        # BGR to RGB 
        frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        # resizing it to feed into model
        frame_resized = cv2.resize(frame_rgb, (width_for_model, height_for_model))
        # input_data will now become 4 dimensional 
        input_data = np.expand_dims(frame_resized, axis=0)
        # now it will have (batchsize, height, width, channel)

        # Normalize pixel values if using a floating model 
        # (i.e. if model is non-quantized)
        if floating_model:
            input_data = (np.float32(input_data) - input_mean) / input_std

        # Perform the actual detection by running the model with the image as input
        boxes, classes, scores = inference.perform_inference(input_data)

        # we could have got  number of objects 
        # but it does not work most of the times.

        # num = interpreter.get_tensor(output_details[3]['index'])[0]  #
        # Total number of detected objects (inaccurate and not needed)

        # The readings array to be inserted in the readings table
        objs = []
        
        # Loop over all detections and draw detection box
        #  if confidence is above minimum then  only 
        #  that detected object will  be considered

        # The index of person class is zero.
        for i in range(len(scores)):
            if (scores[i] > min_conf_threshold) and (int(classes[i] == 0)):
                
                # Get bounding box coordinates and draw box
                # Interpreter can return coordinates that are outside of image dimensions, 
                # need to force them to be within image using max() and min()

                ymin_model = round_to_three_decimal_places(boxes[i][0])
                xmin_model = round_to_three_decimal_places(boxes[i][1])
                ymax_model = round_to_three_decimal_places(boxes[i][2])
                xmax_model = round_to_three_decimal_places(boxes[i][3])

                # map the bounding boxes from model to the window 
                ymin = int(max(1, (ymin_model * camera_width)))
                xmin = int(max(1, (xmin_model * camera_height)))
                ymax = int(min(camera_width, (ymax_model * camera_width)))
                xmax = int(min(camera_height, ( xmax_model * camera_height)))

                # draw the rectangle on the frame
                cv2.rectangle(frame, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)

                # Contructing the label

                # Look up object name from "labels" array using class index
                object_name = labels[int(classes[i])] 

                # Example: 'person: 72%'
                label = '%s: %d%%' % (object_name, int(scores[i]*100)) 

                # Get font size
                labelSize, baseLine = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 
                                                      0.7, 2)
                
                # Make sure not to draw label too close to top of window
                label_ymin = max(ymin, labelSize[1] + 10) 

                # Draw white box to put label text in
                cv2.rectangle(frame, (xmin, label_ymin-labelSize[1]-10), 
                              (xmin+labelSize[0], label_ymin+baseLine-10), 
                              (255, 255, 255), cv2.FILLED)

                # Draw the text label 
                cv2.putText(frame, label, (xmin, label_ymin-7), 
                            cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 2) 
                
                # the readings to be inserted into the table
                objs.append({'label': labels[classes[i]],
                            'score': 100*scores[i],
                            'bounding_box': [xmin, ymin, xmax, ymax]
                            })

        # Draw framerate in corner of frame
        cv2.putText(frame, 'FPS: {0:.2f}'.format(frame_rate_calc),
                    (30, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 0)
                    , 2, cv2.LINE_AA)
        
        # All the results have been drawn on the frame, so it's time to display it.
        global shutdown_in_progress
        if shutdown_in_progress:

            videostream.stop()
            time.sleep(3)
            cv2.destroyWindow(window_name)
            break
        else:

            # Calculate framerate
            t_end = cv2.getTickCount()
            time1 = (t_end-t1)/freq
            frame_rate_calc = 1/time1   

            reads = {}
            for r_index in range(len(objs)):
                reads['person_' + str(r_index+1) + '_' + 'label'] = objs[r_index]['label']
                reads['person_' + str(r_index+1) + '_' + 'score'] = objs[r_index]['score']
                reads['person_' + str(r_index+1) + '_' + 'x1'] = objs[r_index]['bounding_box'][0]
                reads['person_' + str(r_index+1) + '_' + 'y1'] = objs[r_index]['bounding_box'][1]
                reads['person_' + str(r_index+1) + '_' + 'x2'] = objs[r_index]['bounding_box'][2]
                reads['person_' + str(r_index+1) + '_' + 'y2'] = objs[r_index]['bounding_box'][3]
            
            reads['count'] = len(objs)
            data = {
                'asset': asset_name,
                'timestamp': utils.local_timestamp(),
                'key': str(uuid.uuid4()),
                'readings': reads
            }
            if not shutdown_in_progress:
                async_ingest.ingest_callback(c_callback, c_ingest_ref, data)

            # show the frame on the window
            try:
                if show_window :
                    cv2.imshow(window_name, frame)
            except Exception as e:
                _LOGGER.info('exception  {}'.format(e))

            # wait for 1 milli second 
            cv2.waitKey(1)


def plugin_start(handle):
    """ Start the plugin operation - observe video feed from camera device and classify that using the
        configured ML model
        Args:
            handle: handle returned by the plugin initialisation call
        Returns:
            Sends back the reading object (JSON doc) via C/python async ingest interface when it recognizes
            the video frame content
        Raises:
    """
    global loop, async_thread, camera_processing_thread, shutdown_in_progress

    loop = asyncio.new_event_loop()
    try:
        model = handle['model_file']['value']
        labels = handle['labels_file']['value']
        global asset_name
        asset_name = handle['asset_name']['value']
        enable_tpu = handle['enable Edge TPU']['value']
        min_conf_threshold = float(handle['Min Confidence Threshold']['value'])

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)

        with open(labels, 'r') as f:
            pairs = (l.strip().split(maxsplit=1) for l in f.readlines())
            labels = dict((int(k), v) for k, v in pairs)

        # instance of the inference class 
        global inference
        inference = Inference()
        interpreter = inference.get_interpreter(model, enable_tpu, 
                                                labels, min_conf_threshold)
        
        # some extra config parameters required for the camera_loop function

        # Since pixel value can be from 0 to 255 , so we are considering mean to
        # be (0+255)/2 = 127.5 .
        input_mean = 127.5
        input_std = 127.5
        camera_height = 640
        camera_width = 480
        
        source = int(handle['Camera_ID']['value'])

        config_dict = {
            'input_mean': input_mean,
            'input_std': input_std,
            'camera_height': camera_height,
            'camera_width': camera_width,
            'source': source
        }

        def run():
            global loop
            loop.run_forever()

        async_thread = Thread(target=run, name="Async Thread")
        async_thread.start()
        camera_processing_thread = Thread(target=camera_loop, 
                                          name='Camera Processing Thread',
                                          kwargs=config_dict)
        camera_processing_thread.start()

    except Exception as exptn:
        _LOGGER.exception("Human detector plugin  failed to start. Details: %s", str(exptn))
        raise


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
        it should be called when the configuration of the plugin is changed
        during the operation of the south service.
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    
    new_handle = plugin_init(new_config)

    model = new_handle['model_file']['value']
    model = os.path.join(os.path.dirname(__file__), "model", model)

    labels = new_handle['labels_file']['value']
    labels = os.path.join(os.path.dirname(__file__), "model", labels)
    with open(labels, 'r') as f:
        pairs = (l.strip().split(maxsplit=1) for l in f.readlines())
        labels = dict((int(k), v) for k, v in pairs)

    enable_tpu = new_handle['enable Edge TPU']['value']
    min_conf_threshold = float(new_handle['Min Confidence Threshold']['value'])

    global inference, asset_name
    asset_name =  new_handle['asset_name']['value']
    _ = inference.get_interpreter(model, enable_tpu, labels, min_conf_threshold)

    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
        To be called prior to the south service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        None
    Raises:
        None
    """
    global loop, async_thread, camera_processing_thread, shutdown_in_progress 
    try:

        shutdown_in_progress = True
        # allow the stream to stop 
        time.sleep(2)
        
        # stopping  every other thread one by one
        loop.stop()
        loop = None 
        async_thread = None
        camera_processing_thread = None

    except Exception as e:
        _LOGGER.exception(str(e))
        raise


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C/Python async ingest interface
        ingest_ref: C opaque object required to passed back to C/Python async ingest interface
    """
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref


